/*
 * Module Name:     aipsokbps.c
 *
 * Description:     PS/2 keyboard/msr filter driver for Linux on SureOne 4614.
 *
 * (C) Copyright 2005,2007, International Business Machines Corporation.
 * All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * 3.  The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE ORF
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <nobody@us.ibm.com>.
 */


#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 5, 0)
#include <linux/moduleparam.h>
#endif
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/ioctl.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/unistd.h>

#include <linux/time.h>
MODULE_AUTHOR("IBM Retail Store Solutions");
MODULE_DESCRIPTION("IBM SureOne Keyboard Driver");
MODULE_LICENSE("GPL compatible");



/* ****************************************************************
 * local include files
 * ****************************************************************
 */
#include "aipsokbps.h"

#define IBM_4614   7   /* IBM4614 ID */
#define IBM_4613   8   /* IBM4613 ID */

static int machine_type = 0;
module_param(machine_type,int,0);
MODULE_PARM_DESC(machine_type,"Integer value");

/* ****************************************************************
 * constant definitions
 * ****************************************************************
 */
#define SUCCESS             0
#define IBM_POS_KBD_MAJOR   246
#define DEV_NODE_NAME       "aipsokbps"

#ifdef DEBUG
#define DPRINTK(fmt, args...) printk("%s(%s): " fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#define EPRINTK(fmt, args...) printk(KERN_ERR "%s(%s): "     fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)
#define IPRINTK(fmt, args...) printk(KERN_INFO "%s(%s): "    fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)
#define WPRINTK(fmt, args...) printk(KERN_WARNING "%s(%s): " fmt, DEV_NODE_NAME, __FUNCTION__ , ## args)


/* ***************************************************************
 * Function Prototypes
 * ***************************************************************
 */

/* Device */
static int open(struct inode *inode, struct file *file);
static int release(struct inode *inode, struct file *file);
static int read(struct file *file, char *buf, size_t count, loff_t *ppos);
static int write(struct file *file, const char *buf, size_t count, loff_t *ppos);
static int ioctl (struct inode *inode, struct file *fp, unsigned int cmd, unsigned long ioctl_arg);

/* SureOne machine */
static int check_model(void);

/* MSR */
static void msr_init(void);
static void reset_msr_status(void);
static void end_track_timed_out(unsigned long data);
static void inter_track_timed_out(unsigned long data);
static void start_end_track_timer(unsigned long data);
static void stop_end_track_timer(void);
static void start_inter_track_timer(void);
static void stop_inter_track_timer(void);
static int is_valid_single_track_data(unsigned char cmd_reply);
static void __discard_and_deactivate_active_data_buffer(void);
static void discard_and_deactivate_active_data_buffer(void);
static void convert_raw_msr_data(void);
static unsigned char collecting_msr_data(unsigned char scancode);

/* IBM 4613 MRS scancode processing */
static int process_IBM4613_scancodes(unsigned char *pScancode);
static unsigned char set_msrprefix(SET_MSR_COMMAND command, int count);
static unsigned char get_reply_setprefix(unsigned char scancode);

/* Stream Buffer */
static inline void update_stream_buffer(unsigned char scancode);

/* Keybooard / Kbd Controller */
static int kbd_get_ec_level(void);
static unsigned char collecting_ec_level(unsigned char scancode);

/* Filter hooks */
static int pos_scancode_filter(unsigned char *pScancode);
static int kbd_send_data(unsigned char *buf, unsigned char size);


/* ***************************************************************
 * Filter Patch
 * ***************************************************************
 */
struct pc_keyb_filter {
    int (*filter_scancode)(unsigned char *scancode);    /* [IN] */
    int (*write_data)(unsigned char *data, unsigned int length, /* [OUT] */
                      unsigned int *xferred, unsigned int retries,
                      unsigned int timeout);
};

extern int  register_pc_keyb_filter(struct pc_keyb_filter *filter);
extern void unregister_pc_keyb_filter(struct pc_keyb_filter *filter);

/* Pass the scancode filter function name.  Upon return
 * pos_filter.write_data() will contain the address of the
 * actual write funtion supported by atkbd/pc_keyb driver.
 */
static struct pc_keyb_filter pos_filter =
{
    filter_scancode:  pos_scancode_filter,
};


/* ***************************************************************
 * Globals
 * ***************************************************************
 */
/* Our major number - may also be set as a module parameter */
static int aipsokbps_MajorNumber = IBM_POS_KBD_MAJOR;

/* Flag indicating we sent a command to the keyboard and it is not
 * yet completed.
 */
static int cmd_busy = FALSE;

/* Main structure that holds all keyboard data. */
static KEYBOARD_DATA  KbdData;
static PKEYBOARD_DATA pKbdData;

/* Machine Determination */
static unsigned long model = 0x00000000;

/* Scancode -> Ascii conversion buffer for MSR data */
static SUREONE_ASCII_BUFFER AsciiBuffer;


/* ***************************************************************
 * File I/O
 * ***************************************************************
 */
struct file_operations sokb_fops = 
{
    .read     = read,
    .write    = write,
    .ioctl    = ioctl,
    .open     = open,
    .release  = release,
};



DECLARE_WAIT_QUEUE_HEAD(wq);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
MODULE_PARM(aipsokbps_MajorNumber, "i");
#else
module_param_named(major, aipsokbps_MajorNumber, int, 0644);
#endif


/************************************************************************
 *
 * Function Name:  sokb_init()
 *
 * Purpose:        Intialization routine
 *
 * Description:    entry point for this driver.
 *
 *
 * Output:
 *                 0        - success
 *                 non-zero - failure.
 *
 * ***********************************************************************
 */
static int __init sokb_init(void)
{
    int rc = SUCCESS;
    int cmd_retries = 0;
    int resend_retries = 0;
    int state = 0;
    int count = 0;
    int rcvd02 = 0;
    int rcvd82 = 0;

    if (machine_type == IBM_4613)
        IPRINTK("Found IBM 4613 Anyplace Kiosk machine\n");
    else if (machine_type == IBM_4614  && check_model() == TRUE && IS_SUREONE(model))
        IPRINTK("Found IBM 4614 SureOne Machine\n");
    else {
        EPRINTK("This driver only supports either SureOne or IBM 4613\n"); 
        return -ENOSYS;
    }

    /* Initialize keyboard data structure */
    memset(&KbdData, 0, sizeof(KEYBOARD_DATA));
    pKbdData = &KbdData;


    /* Register with low level driver, atkbd / pc_keyb driver. Upon
     * registration, the atkbd / pc_keyb driver provides a pointer to write
     * function.  In addtion, it calls our interrupt routine
     * (pos_scancode_filter()) each time keyboard interrupts.
     * ... this should be OK if the patch for atkbd/pc_keyb is applied.
     * However, we need to find a way to check for the existence of
     * the symbols before calling them
     */
    DPRINTK("Registering with keyboard driver - atkbd\n");
    register_pc_keyb_filter(&pos_filter);

    if (machine_type == IBM_4614) {
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(10);

        /* Read Keyboard Controller EC Level */
        rc = kbd_get_ec_level();

        if (rc != 0) {
            EPRINTK("kbd_get_ec_level failed rc=%02x\n", rc);

            /* Unregister with atkb/pc_keyb driver */
            unregister_pc_keyb_filter(&pos_filter);
            return -ENODEV;
        }

        /* Wait until we have received keyboard controller EC byte
         * max wait = 5*10(ms)*20 = 1000 ms, timer tick = 10ms
         */
        cmd_retries = 20;
        while (cmd_busy && cmd_retries) {
            DPRINTK(" ... cmd_busy = %02x\n", cmd_busy);
            set_current_state(TASK_INTERRUPTIBLE);
            schedule_timeout(5);
            --cmd_retries;
        }

        if (pKbdData->EC_Level != 0) {
            IPRINTK("Found SureOne keyboard controller, EC Level %d\n",
                    pKbdData->EC_Level);
        } else {
            /* Did not find a Keyboard Controller */
            EPRINTK("No SureOne keyboard controller found\n");

            /* Unregister with atkbd/pc_keyb driver */
            unregister_pc_keyb_filter(&pos_filter);
            return -ENODEV;
        }

        DPRINTK("aipsokbps.c :  _init() -->Initializing MSR\n");
    }

    if (machine_type == IBM_4613){

       pKbdData->fStatus |= SET_MSR_PREFIX;
	   for ( count = 0; count <= 2; count++){
		// For first command
		   for ( state = 1; state <= 4; state++){
		     set_current_state(TASK_INTERRUPTIBLE);
                     schedule_timeout(10);
		     rc = set_msrprefix(state,count);
		     if (rc != 0) {
                        EPRINTK("set_msrprefix failed rc=%02x\n", rc);
                       /* Unregister with atkb/pc_keyb driver */
                       unregister_pc_keyb_filter(&pos_filter);
                       rc = -ENODEV;
		       break;
		     }
	     
             	/* Wait until we have received keyboard controller EC byte
              	 * max wait = 5*10(ms)*20 = 1000 ms, timer tick = 10ms
             	 */
             	cmd_retries = 5;
             	while (cmd_busy && cmd_retries) {
                DPRINTK(" ... cmd_busy = %02x\n", cmd_busy);
                set_current_state(TASK_INTERRUPTIBLE);
                schedule_timeout(5);
                --cmd_retries;
		}

             	if( state == 4){
                      
                	if(pKbdData->mStatus == SCANCODE_ACK){
			DPRINTK("ACK Received\n");
                 	/*Wait for the next byte*/
		          cmd_retries = 5;
		          while (cmd_busy && cmd_retries) {
			    DPRINTK(" ... cmd_busy = %02x\n", cmd_busy);
			    set_current_state(TASK_INTERRUPTIBLE);
			    schedule_timeout(5);
			   --cmd_retries;

		          }
                      }
                      if(pKbdData->mStatus == SCANCODE_ONE ){
                          rcvd02 = 1;
                      }
                      else if (pKbdData->mStatus == 
                               BREAK_SCANCODE(SCANCODE_ONE)){
                          rcvd82 = 1;
                      }

                      /*Wait for the next byte*/
		      cmd_retries = 5;
		      while (cmd_busy && cmd_retries) {
			DPRINTK(" ... cmd_busy = %02x\n", cmd_busy);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5);
			--cmd_retries;
		      }

                      if(pKbdData->mStatus == SCANCODE_ONE ){
                          rcvd02 = 1;
                      }
                      else if (pKbdData->mStatus == 
                         BREAK_SCANCODE(SCANCODE_ONE)){
                          rcvd82 = 1;
                      }
                      
                      if( rcvd02 == 0 || rcvd82 == 0 )
                      {
                        /*wait last time for the reply pair*/
                        /*Wait for the next byte*/
		        cmd_retries = 5;
		        while (cmd_busy && cmd_retries) {
			DPRINTK(" ... cmd_busy = %02x\n", cmd_busy);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5);
			--cmd_retries;
		        }
                      }
                   }
		 else{
			if(pKbdData->mStatus == SCANCODE_ACK){
				DPRINTK("ACK Received\n");
			 }
			else if ( pKbdData->mStatus == SCANCODE_RESEND){
				if ( resend_retries == 5 ){
					DPRINTK("BAIL OUT..WE HAVE BEEN NACKED 5 times\n");
					unregister_pc_keyb_filter(&pos_filter);
					rc = -ENODEV;
					break;
				}
				DPRINTK("NACK RECEIVED?->0x%02x then Retry\n",pKbdData->mStatus);
				state--;
				resend_retries++;
			}
			else{
				DPRINTK("Retry -> UNKNOWN REPLY RECEIVED?->0x%02x\n",pKbdData->mStatus);
				if ( resend_retries == 5 ){
					DPRINTK("BAIL OUT..WE HAVE TRIED LONG ENOUGH\n");
					unregister_pc_keyb_filter(&pos_filter);
					rc = -ENODEV;
					break;
				}
				state--;
				resend_retries++;
			}
		 }
	}
	if ( rc != 0 )
		break;
	}
    pKbdData->fStatus &= ~SET_MSR_PREFIX;
}

	if ( rc != 0 )
	  return rc;
    /* Initialise MSR variables and objects */
    msr_init();


    /* Register driver */
    rc = register_chrdev(aipsokbps_MajorNumber,   /*  major number */
                         DEV_NODE_NAME,           /*  node name */
                         &sokb_fops);             /*  file ops */

    if (rc != 0) {
        EPRINTK("Register device with kernel failed rc=0x%02x\n", rc);

        /* Unregister with atkbd/pc_keyb driver */
        unregister_pc_keyb_filter(&pos_filter);
        return -ENODEV;
    }

    IPRINTK("Loaded, MajorNumber=%d\n", aipsokbps_MajorNumber);

    return rc;
}



/************************************************************************
 *
 * Function Name:  sokb_exit()
 *
 * Purpose:        driver exit routine.
 *
 * Description:    This routine gets called when driver unloads.
 *
 *
 * Output:
 *                 none
 *
 * ***********************************************************************
 */
static void __exit sokb_exit(void)
{
    int rc = SUCCESS;

    /* Unregister driver with the kernel */
    rc = unregister_chrdev(aipsokbps_MajorNumber, DEV_NODE_NAME);

    if (rc != 0) {
        EPRINTK("Unregister device with kernel failed rc=0x%02x\n", rc);
    }
#ifdef DEBUG
    else {
        DPRINTK("Unregistered device with kernel\n");
    }
#endif

    /* Unregister with atkbd / pc_keyb driver. */
    unregister_pc_keyb_filter(&pos_filter);
}



/************************************************************************
 *
 * Function Name:  open()
 *
 * Purpose:        driver open routine.
 *
 * Description:    This routine gets called when application opens this
 *                 driver.
 *
 *
 * Output:
 *                 0     -  success
 *
 *
 * ***********************************************************************
 */
static int open(struct inode *inode, struct file *file)
{
    DPRINTK("MAJOR=%d, MINOR=%d\n",
            MAJOR (inode->i_rdev), MINOR (inode->i_rdev));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_INC_USE_COUNT;
#endif

    return SUCCESS;
}



/************************************************************************
 *
 * Function Name:  release()
 *
 * Purpose:        driver close routine.
 *
 * Description:    This routine gets called when application closes this
 *                 driver.
 *
 * Output:
 *                 0  -  success
 *
 * ***********************************************************************
 */
static int release (struct inode *inode, struct file *file)
{
    DPRINTK("closing character device: %s\n", DEV_NODE_NAME);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_DEC_USE_COUNT;
#endif

    return SUCCESS;
}



/************************************************************************
 *
 * Function Name:  read()
 *
 * Purpose:        driver read routine.
 *
 * Description:    This routine gets called when application sends a
 *                 read request.
 *
 * Output:
 *                 EINVAL
 *
 * Notes:          read request is not supported.
 *
 * ***********************************************************************
 */
static int read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    return -EINVAL;
}



/************************************************************************
 *
 * Function Name:  write()
 *
 * Purpose:        driver write routine.
 *
 * Description:    This routine gets called when application sends a
 *                 write request.
 *
 * Output:
 *                 EINVAL
 *
 * Notes:          write request is not supported.
 *
 * ***********************************************************************
 */
static int write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    return -EINVAL;
}



/************************************************************************
 *
 * Function Name:  kbd_send_data()
 *
 * Purpose:        To send data to keyboard.
 *
 * Description:    A common routine to send data to keyboard.
 *
 * Input:          unsigned char * buffer.
 *                 size of buffer
 *
 * Output:
 *                 0 - success
 *                 1 - failure.
 *
 * ***********************************************************************
 */
static int kbd_send_data(unsigned char *buf, unsigned char size)
{
    int rc = 0;
    int cmd_retries = 100;
    int xferred = 0, xfer_retry = size;
    unsigned char b = 0;

    if (size > 1)
        b = *(buf + 1);

    DPRINTK("enter\n");

    while (cmd_busy && cmd_retries) {
        DPRINTK("going to sleep ...\n");
        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(5);
        --cmd_retries;
        DPRINTK("... return from sleep\n");
    }

    /* Call the Linux kbd driver's send routine to send data. */

    cmd_busy = 1;

    do {
        DPRINTK("sending %d bytes to kbd\n", size);
        DPRINTK("pos_filter.write_data =  %p\n", pos_filter.write_data);
        /* size: number of retires and size of buffer */
        rc = pos_filter.write_data(buf, size, &xferred, size,1000);

        if (rc != 0) {
            DPRINTK("pos_filter.write_data returned error %d\n", rc);

            /* The kernel patch misses ACK, and does
             * not increment the xferred count
             */
            if (rc == (-EINTR))
                xferred++;

            if (xferred < size) {
                buf += xferred;
                size -= xferred;
                xferred = 0;
            }
        }
    }
    while (xferred < size && xfer_retry--);

    /* Go to sleep
     * The wait queue is awakened from the pos_kbd_filter() routine when the 
     * command is complete
     */
    if (rc != 0)
        cmd_busy = 0;

    if (cmd_busy != 0)
        interruptible_sleep_on_timeout(&wq, 100);

    DPRINTK("exit, rc=0x%02x %s\n", rc, rc == 0 ? "Success" : "Failed");

    return rc;
}



/************************************************************************
 *
 * Function Name:  kbd_get_ec_level()
 *
 * Purpose:        To read Keyboard Controller EC level
 *
 * Description:    This routine sends 0xEB, Read Keyboard Controller Firmware
 *                 Level command. The keyboard response contains one byte.
 *
 * Input:          none
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */
static int kbd_get_ec_level(void)
{
    int rc;
    unsigned char inputBuffer[1];

    DPRINTK("enter\n");

    /* Set up EC Level Command */
    inputBuffer[0] = 0xEB;

    /* Set global state */
    pKbdData->fStatus |= COLLECTING_EC_LEVEL;

    /* Write to keyboard */
    rc = kbd_send_data(inputBuffer, sizeof(inputBuffer));

    DPRINTK("exit, rc=0x%02x %s\n", rc, rc == 0 ? "Success" : "Failed");

    return rc;
}



/************************************************************************
 *
 * Function Name:  collecting_ec_level()
 *
 * Purpose:        To read Keyboard Controller EC level
 *
 * Description:    This routine collects the response to the 0xEB Read
 *                 Keyboard Controller Firmware Level command. The keyboard
 *                 response contains one byte.
 *
 * Input:          scancode
 *
 * Output:         none
 *
 * Notes:
 *
 * ***********************************************************************
 */
static unsigned char collecting_ec_level(unsigned char scancode)
{
    DPRINTK("scancode=0x%02x\n", scancode);

    pKbdData->EC_Level = scancode;

    pKbdData->fStatus &= ~COLLECTING_EC_LEVEL;

    DPRINTK("EC Level %d\n", pKbdData->EC_Level);

    /* Wake up  */
    cmd_busy = 0;
    wake_up_interruptible(&wq);

    scancode = 0;

    return scancode;
}



/************************************************************************
 *
 * Function Name:  get_reply_setprefix()
 *
 * Purpose:        To read Keyboard Controller Status Reply
 *
 * Description:    This routine collects the response to the commands on
 *                 setting the prefix,suffix and enabling the feature.
 *				   The keyboard response contains one byte.
 *
 * Input:          scancode
 *
 * Output:         scancode
 *
 * Notes:
 *
 * ***********************************************************************
 */
static unsigned char get_reply_setprefix(unsigned char scancode)
{
    DPRINTK("scancode=0x%02x\n", scancode);

    pKbdData->mStatus = scancode;

    DPRINTK("Status reply %d\n", pKbdData->mStatus);

    /* Wake up  */
    cmd_busy = 0;
    wake_up_interruptible(&wq);

    scancode = 0;

    return scancode;
}
/*********************************************************************
 *
 * Function Name: collecting_msr_data
 *
 * Purpose:       Save MSR data (status) scancodes
 *
 * Description:   This function saves the MSR data scancodes in the
 *                MSRData buffer.  The the scancode is set to 0
 *
 * Input:         scancode - the scancode received from the keyboard
 *
 * Output:        
 *                0
 *
 * ********************************************************************
 */
static unsigned char collecting_msr_data(unsigned char scancode)
{
    DPRINTK("collecting_msr_data scancode=0x%02x\n", scancode);

    /* Collect and convert using mapping table */
    if ( scancode <= 127 )
    pKbdData->MSRData.MSRDataBuffer[pKbdData->MSRData.Index] = convert[scancode];
    else
      pKbdData->MSRData.MSRDataBuffer[pKbdData->MSRData.Index] = scancode;

    DPRINTK("convert[scancode]=0x%02x\n", pKbdData->MSRData.MSRDataBuffer[pKbdData->MSRData.Index]);
    scancode = 0; /* Consumed */

    pKbdData->MSRData.Index += 1;
    pKbdData->MSRData.MSRDataLength += 1;

    return scancode;
}


/*********************************************
* FunctionName: set_msrprefix
* Purpose     : send command to keyboard controller
*               to set msr prefix/suffix
* Input:      : command: 
*               count: 
                      0 : prefix
                      1 : suffix
                      2 : enable prefix and suffix
*
* Output: status of the last command sent
*
*********************************************/
static unsigned char set_msrprefix(SET_MSR_COMMAND command, int count)
{

    int rc;
    unsigned char inputBuffer[1];

    DPRINTK(" set_msrprefix enter\n");

	switch ( command ){
		case FirstMsrPrefixCommand:{
			inputBuffer[0] = COMMAND1;
			break;
		}
		case SecondMsrPrefixCommand:{
			inputBuffer[0] = COMMAND2;
			break;
		}
		case LastMsrPrefixCommand:{
			if ( count == 0 )
				inputBuffer[0] = SETPREFIX;
			else if ( count == 1 )
				inputBuffer[0] = SETSUFFIX;
			else if ( count == 2 )
				inputBuffer[0] = ENABLE_PREF_AND_SUFFIX;
			break;
		}
		case SetPrefixAndSuffix:{
			if ( count == 0 )
				inputBuffer[0] = PREFIX_VAL;
                        else if ( count == 1)
				inputBuffer[0] = SUFFIX_VAL;
			else if ( count == 2)
				inputBuffer[0] = ENABLE_PREF_AND_SUFFIX_VAL;
			break;
		}
		default:
			DPRINTK("Unknown command\n");
	}
	

    /* Write to keyboard */
    rc = kbd_send_data(inputBuffer, sizeof(inputBuffer));

    DPRINTK("exit, rc=0x%02x %s\n", rc, rc == 0 ? "Success" : "Failed");

    return rc;
}





/*********************************************************************
 * Function Name: process_IBM4613_scancodes
 * Purpose:       Filterout IBM4613-MSR specific data
 * Description:   This function is another branch of pos_scancode_filter 
 *                which filters only IBM4613 specific scancode processing
 * 
 * Input:         scancode sent for latest keyboard interrupt
 *
 * Output:        0 - Ask the Linux kbd driver interrupt routine to 
 *                    continue processing this scancode
 *
 *                1 - Ask the Linux kbd driver interrupt routine to
 *                    ignore this scancode and return immediately as
 *                    we have filtered it off (consumed it)
 *
 *********************************************************************
 */
static int process_IBM4613_scancodes(unsigned char *pScancode)
{
    unsigned char scancode = *pScancode;
    int rc = 0;
    
    int x = 0;

    SUREONE_MSR_COLLECTION_STATE msrState = pKbdData->SureOneMsrStatus.CollectionState;

    /***********************************
    *  Check for setting prefix and suffix
    *
    ************************************/
    if ( pKbdData->fStatus == SET_MSR_PREFIX || pKbdData->fStatus ==  SET_MSR_SUFFIX ){ 
		
		scancode = get_reply_setprefix(scancode);
                *pScancode = scancode = 0;
                return 1;

   }
    else if (pKbdData->fStatus == COLLECTING_MSR_DATA) {
        stop_end_track_timer();
        switch (scancode) {
        case RALT:
        {
            pKbdData->SureOneMsrStatus.CollectionState = POST_FIX_DETECTED;
            *pScancode = scancode = 0;
            return 1;

        }
        case KEYPAD1:
            {
                if (pKbdData->StreamBuffer[1] == RALT) {
                    pKbdData->SureOneMsrStatus.CollectionState = POST_FIX_DETECTED;

                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                   // DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                } else
                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;

                *pScancode = scancode = 0;
                return 1;


            }
        case BREAK_SCANCODE(KEYPAD1):
            {
                if (pKbdData->SureOneMsrStatus.CollectionState == POST_FIX_DETECTED &&
                    pKbdData->StreamBuffer[2] == RALT &&
                    pKbdData->StreamBuffer[1] == KEYPAD1) {
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                    //DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode = 0;
                    return 1;
                } else {
                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;
                    __discard_and_deactivate_active_data_buffer();
                    *pScancode = scancode;
                    return 0;
                }



            }
        case KEYPAD2:
            {
                if (pKbdData->SureOneMsrStatus.CollectionState == POST_FIX_DETECTED &&
                    pKbdData->StreamBuffer[3]== RALT &&
                    pKbdData->StreamBuffer[2]== KEYPAD1 &&
                    pKbdData->StreamBuffer[1] == BREAK_SCANCODE(KEYPAD1)) {
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                    //DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                } else
                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;

                *pScancode = scancode = 0;
                return 1;

            }
        case BREAK_SCANCODE(KEYPAD2):
            {
                if (pKbdData->SureOneMsrStatus.CollectionState == POST_FIX_DETECTED &&
                    pKbdData->StreamBuffer[4] == RALT &&
                    pKbdData->StreamBuffer[3] == KEYPAD1 &&
                    pKbdData->StreamBuffer[2] == BREAK_SCANCODE(KEYPAD1) &&
                    pKbdData->StreamBuffer[1] == KEYPAD2) {
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                    //DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode = 0;
                    return 1;
                } else {

                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;
                    __discard_and_deactivate_active_data_buffer();
                    *pScancode = scancode;
                    return 0;
                }                       

            }
        case KEYPAD6:
            {
                if (pKbdData->SureOneMsrStatus.CollectionState == POST_FIX_DETECTED &&
                    pKbdData->StreamBuffer[5] == RALT &&
                    pKbdData->StreamBuffer[4] == KEYPAD1 &&
                    pKbdData->StreamBuffer[3] == BREAK_SCANCODE(KEYPAD1) &&
                    pKbdData->StreamBuffer[2] == KEYPAD2 &&
                    pKbdData->StreamBuffer[1] == BREAK_SCANCODE(KEYPAD2)) {
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                    DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode = 0;
                    return 1;

                } else {
                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;
                    __discard_and_deactivate_active_data_buffer();
                    DPRINTK("DISCARD AND DEACTIVATE NOT EXPECTED<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode;
                    return 0;
                }

            }
        case BREAK_SCANCODE(KEYPAD6):
            {

                if (pKbdData->SureOneMsrStatus.CollectionState == POST_FIX_DETECTED &&
                    pKbdData->StreamBuffer[6] == RALT &&
                    pKbdData->StreamBuffer[5] == KEYPAD1 &&
                    pKbdData->StreamBuffer[4] == BREAK_SCANCODE(KEYPAD1) &&
                    pKbdData->StreamBuffer[3] == KEYPAD2 &&
                    pKbdData->StreamBuffer[2] == BREAK_SCANCODE(KEYPAD2) &&
                    pKbdData->StreamBuffer[1] == KEYPAD6) {
                    DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    stop_end_track_timer();
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                    *pScancode = scancode = 0;
                    return 1;
                } else {
                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;
                    __discard_and_deactivate_active_data_buffer();
                    DPRINTK("DISCARD AND DEACTIVATE NOT EXPECTED<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode;
                    return 0;
                }



            }
        case BREAK_SCANCODE(RALT):
            {
                if (pKbdData->SureOneMsrStatus.CollectionState == POST_FIX_DETECTED &&
                    pKbdData->StreamBuffer[6] == KEYPAD1 &&
                    pKbdData->StreamBuffer[5] == BREAK_SCANCODE(KEYPAD1) &&
                    pKbdData->StreamBuffer[4] == KEYPAD2 &&
                    pKbdData->StreamBuffer[3] == BREAK_SCANCODE(KEYPAD2) &&
                    pKbdData->StreamBuffer[2] == KEYPAD6 &&
                    pKbdData->StreamBuffer[1] == BREAK_SCANCODE(KEYPAD6)) {

                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;
                    DPRINTK("<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode;

                    /* Switch SureOne MSR Status to 'Holding MSR Data'
                     * This locks the buffer
                     */
                    DPRINTK("    Moving to Collection State: Inactive\n");
                    pKbdData->fStatus |= Inactive;
                    pKbdData->SureOneMsrStatus.CollectionState = Inactive;

                    				
			#ifdef DEBUG
			DPRINTK("COMPLETE DATA IN BUFFER:\n");
			for ( x = 0; x < pKbdData->MSRData.MSRDataLength ; x++ )
				printk("0x%02x ", pKbdData->MSRData.MSRDataBuffer[x]);
				printk("\n");
			#endif

                    /* Switch Off collecting phase */
                    DPRINTK("    Switching Off COLLECTING_MSR_DATA\n");
                    return 0;


                } else {
                    pKbdData->fStatus &= ~COLLECTING_MSR_DATA;
                    __discard_and_deactivate_active_data_buffer();
                    DPRINTK("DISCARD AND DEACTIVATE NOT EXPECTED<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                    *pScancode = scancode;
                    return 0;
                }

            }

        default:
            {
                /*discard Break scancodes since it causes double data entry to app*/
                /*if ( scancode != BREAK_SCANCODE(pKbdData->StreamBuffer[1]))  */            if ( scancode < BREAK ||
		 scancode == BREAK_SCANCODE(SCANCODE_LEFT_SHIFT)){
                  scancode = collecting_msr_data(scancode);
                  DPRINTK("COLLECTING MSR DATA<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                }
                else
                   DPRINTK("Break Scancode: DONT COLLECT\n");
                 //retain integrity of Status
                pKbdData->fStatus &= ~pKbdData->fStatus;
                pKbdData->fStatus |= COLLECTING_MSR_DATA;
                *pScancode = scancode = 0;
                return 1;
            }

        }

    } 
	else {
        if (msrState == Inactive && scancode == RALT) {
            /* catch here the scenario where some cards does not report
             * prefixes
             */
                    /* This will be true if there were no PREFIX,so this
                     * will be the suffix part
                     */
                        *pScancode = scancode;
                        pKbdData->SureOneMsrStatus.CollectionState = Waiting;
                        start_end_track_timer((unsigned long) pKbdData->SureOneMsrStatus.CollectionState);
     

		}
       else if ( msrState == Waiting ){
            DPRINTK("MSR WAITING...PROCESSING SCANCODE\n");       
            switch (scancode){
			case KEYPAD1:
                {
		    stop_end_track_timer();
                    if (pKbdData->StreamBuffer[1] == RALT) {
                        /*clear buffer for the uncollected data stream*/
                         __discard_and_deactivate_active_data_buffer();
                         
                         /*resume collection state*/
                         pKbdData->SureOneMsrStatus.CollectionState = Waiting;

                        *pScancode = scancode = 0;
						
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        return 1;
                    }
                    __discard_and_deactivate_active_data_buffer();
                    break;

                }
            case BREAK_SCANCODE(KEYPAD1):
                {
                    stop_end_track_timer();

                    if (pKbdData->StreamBuffer[2] == RALT &&
                        pKbdData->StreamBuffer[1] == KEYPAD1) {
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        *pScancode = scancode = 0;
                        return 1;
                    } else {
                        /* return, this is a BREAK/release scancode */
                        __discard_and_deactivate_active_data_buffer();
                        *pScancode = scancode;
                        return 0;
                    }
                    break;
                }
            case KEYPAD2:
                {
                    stop_end_track_timer();

                    if (pKbdData->StreamBuffer[3]== RALT &&
                        pKbdData->StreamBuffer[2]== KEYPAD1 &&
                        pKbdData->StreamBuffer[1] == BREAK_SCANCODE(KEYPAD1)) {
                        *pScancode = scancode = 0;
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        return 1;
                    }
                    __discard_and_deactivate_active_data_buffer();
                    break;
                }
            case BREAK_SCANCODE(KEYPAD2):
                {
                    stop_end_track_timer();

                    if (pKbdData->StreamBuffer[4] == RALT &&
                        pKbdData->StreamBuffer[3] == KEYPAD1 &&
                        pKbdData->StreamBuffer[2] == BREAK_SCANCODE(KEYPAD1) &&
                        pKbdData->StreamBuffer[1] == KEYPAD2) {
                        *pScancode = scancode = 0;
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        return 1;
                    } else {
                        /* return; this is a BREAK/release scancode */
                        __discard_and_deactivate_active_data_buffer();
                        *pScancode = scancode;
                        return 0;
                    }
                    break;
                }
            case KEYPAD7:
                {
                    if (pKbdData->StreamBuffer[5] == RALT &&
                        pKbdData->StreamBuffer[4] == KEYPAD1 &&
                        pKbdData->StreamBuffer[3] == BREAK_SCANCODE(KEYPAD1) &&
                        pKbdData->StreamBuffer[2] == KEYPAD2 &&
                        pKbdData->StreamBuffer[1] == BREAK_SCANCODE(KEYPAD2)) {
                        *pScancode = scancode = 0;
                        stop_end_track_timer();
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        return 1;
                    }
                    __discard_and_deactivate_active_data_buffer();
                    break;
                }
            case BREAK_SCANCODE(KEYPAD7):
                {
                    if (pKbdData->StreamBuffer[6] == RALT &&
                        pKbdData->StreamBuffer[5] == KEYPAD1 &&
                        pKbdData->StreamBuffer[4] == BREAK_SCANCODE(KEYPAD1) &&
                        pKbdData->StreamBuffer[3] == KEYPAD2 &&
                        pKbdData->StreamBuffer[2] == BREAK_SCANCODE(KEYPAD2) &&
                        pKbdData->StreamBuffer[1] == KEYPAD7) {
                        *pScancode = scancode = 0;
                        stop_end_track_timer();
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        return 1;
                    }
                    __discard_and_deactivate_active_data_buffer();
                    break;
                }
            case BREAK_SCANCODE(RALT):
                {
                    stop_end_track_timer();

                    if (pKbdData->StreamBuffer[6] == KEYPAD1 &&
                        pKbdData->StreamBuffer[5] == BREAK_SCANCODE(KEYPAD1) &&
                        pKbdData->StreamBuffer[4] == KEYPAD2 &&
                        pKbdData->StreamBuffer[3] == BREAK_SCANCODE(KEYPAD2) &&
                        pKbdData->StreamBuffer[2] == KEYPAD7 &&
                        pKbdData->StreamBuffer[1] == BREAK_SCANCODE(KEYPAD7)) {
						pKbdData->fStatus &= ~Waiting;
                        pKbdData->fStatus |= COLLECTING_MSR_DATA;// CHECKING_NEXT_SCANCODE;
                        pKbdData->SureOneMsrStatus.CollectionState &= ~Waiting;
                        pKbdData->SureOneMsrStatus.CollectionState |= COLLECTING_MSR_DATA;
                        DPRINTK("WILL CHECK FOR NEXT SCANCODE<<<-  rc=0x01, scancode=0x%02x\n",scancode);
                        start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);
                        *pScancode = scancode = 0; /*return the break*/
                        return 0;
                    }
                    __discard_and_deactivate_active_data_buffer();
                    break;
                }

            default:
                break;
            }
        }
        else{
            DPRINTK("Keyboard Data = 0x%02x\n",scancode);
           *pScancode = scancode;
        }
    }

    if (scancode == 0) {
        *pScancode = 0;
        rc = 1;
    } else {
        *pScancode = scancode;
        rc = 0;
    }
    DPRINTK("EXIT <<<- rc= 0x%2x, scancode = 0x%2x\n",rc,*pScancode);
    return rc;
}


/*********************************************************************
 *
 * Function Name: pos_scancode_filter
 *
 * Purpose:       Filter out the MSR specific data
 *
 * Description:   This function looks for specific data (like MSR and
 *                EC Level) and filters it out before the normal keyboard
 *                interrupt routine, and thus the linux keyboard driver
 *                stack, gets a chance to process it.
 *
 * Input:         scancode sent for latest keyboard interrupt
 *
 * Output:        0 - Ask the Linux kbd driver interrupt routine to 
 *                    continue processing this scancode
 *
 *                1 - Ask the Linux kbd driver interrupt routine to
 *                    ignore this scancode and return immediately as
 *                    we have filtered it off (consumed it)
 *
 * ********************************************************************
 */

static int pos_scancode_filter(unsigned char *pScancode)
{
    unsigned char scancode = *pScancode;
    int rc = 0;

    /* Track latest few scancodes in the stream */
    update_stream_buffer(scancode);

   // DPRINTK("aipsokbps.c  pos_scancode_filter()->>> scancode=0x%02x, *pScan=0x%02x, pKbdData->fStatus=0x%02x\n", scancode, *pScancode, pKbdData->fStatus);

    if (machine_type == IBM_4613) {
        rc = process_IBM4613_scancodes(pScancode);
        return rc;
    } else if (machine_type == IBM_4614) {
        /* //////////////////////
         * EC Level - cmd 0xEB //
         * //////////////////////
         */
        if ((pKbdData->fStatus & COLLECTING_EC_LEVEL) &&
            (scancode != SCANCODE_ACK) && (scancode != 0x9c)) {
            scancode = collecting_ec_level(scancode);

           // DPRINTK("COLLECTING EC LEVEL <<<-  rc=0x01, scancode=0x00\n");
            *pScancode = scancode = 0;
            return 1;
        }

        /* //////
         * MSR //
         * //////
         */
        else if (pKbdData->fStatus & CHECKING_FOR_MSR_SS_27_ACK) {
            DPRINTK("Current Status is: CHECKING_FOR_MSR_SS_27_ACK\n");
            DPRINTK("    Command 0xE5 returns: 0x%02x\n", scancode);

            switch (scancode) {
            /* A missing Ack is the `broken` behaviour the SureOne keyboard
             * controller.
             *
             * If MSR is active it sends a Nack 0xFE plus a bad status byte
             * 0x55.
             *
             * Linux 2.4:
             *   The bad status means that the scancode is ignored by linux
             *   in pc_keyb.c
             *  (handle_kbd_event) and never reaches the driver.
             *
             * Linux 2.6:
             *   0xFE comes to this filter
             */
            
            case 0xFA:  /* Ack for Command E5, implies Source is keyboard */
                {
                    DPRINTK("    Switching Off CHECKING_FOR_MSR_SS_27_ACK\n");

                    pKbdData->fStatus &= ~CHECKING_FOR_MSR_SS_27_ACK;

                    DPRINTK("    Switching On CHECKING_SOURCE_KEYBOARD_27\n");

                    pKbdData->fStatus |= CHECKING_SOURCE_KEYBOARD_27;

                    DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                    *pScancode = scancode = 0;
                    return 1;
                }
            default:    /* No Ack for Command E5, implies Source is MSR */
                {
                    DPRINTK("aipsokbps.c No ACK for Command E5, implies source is MSR\n");
                    /* Set SureOneMsrStatus collection stage */
                    if (pKbdData->SureOneMsrStatus.CollectionState == Waiting) {
                        DPRINTK("    Moving to Collection State: got_SS_1\n");

                        pKbdData->SureOneMsrStatus.CollectionState = got_SS_1;
                    } else if (pKbdData->SureOneMsrStatus.CollectionState == FirstTrackComplete) {
                        DPRINTK("    Moving to Collection State: got_SS_2\n");

                        pKbdData->SureOneMsrStatus.CollectionState = got_SS_2;

                        /* Switch Off the safety timer between
                         * FirstTrackComplete and SS 2
                         */
                        stop_inter_track_timer();
                    } else {
                        EPRINTK("    Unexpected Collection state 0x%02x\n",
                                pKbdData->SureOneMsrStatus.CollectionState);

                        /* Discard the buffer and Deactivate
                         * This also switches Off all Flags and puts us back
                         * into Inactive state
                         */
                        EPRINTK("    Discarding MSR buffer and deactivating\n");                  

                        __discard_and_deactivate_active_data_buffer();

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }

                    DPRINTK("    Switching Off CHECKING_FOR_MSR_SS_27_ACK\n");

                    pKbdData->fStatus &= ~CHECKING_FOR_MSR_SS_27_ACK;

                    DPRINTK("    Switching On COLLECTING_MSR_DATA\n");

                    pKbdData->fStatus |= COLLECTING_MSR_DATA;

                    /* Start safety timeout timer: SS->End of Track */
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);

                    /* Change scancode to start sentinel
                     * We want to save the sentinel as part of the MSR data
                     * for SureOne
                     */
                    scancode = collecting_msr_data(0x27);

                    DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                    *pScancode = scancode = 0;
                    return 1;
                }
            }
        } else if (pKbdData->fStatus & CHECKING_SOURCE_KEYBOARD_27) {
            DPRINTK("Current Status is: CHECKING_SOURCE_KEYBOARD_27\n");
            DPRINTK("    Command 0xE5 Reply returns: 0x%02x\n", scancode);
            DPRINTK("    Switching Off CHECKING_SOURCE_KEYBOARD_27\n");

            pKbdData->fStatus &= ~CHECKING_SOURCE_KEYBOARD_27;

            /* Change scancode to ';' Keypress
             * We want to pass on the original key
             */
            scancode = 0x27;

            DPRINTK("<<<-  rc=0x00, scancode=0x27\n");
            *pScancode = scancode;
            return 0;
        } else if (pKbdData->fStatus & CHECKING_FOR_MSR_SS_06_ACK) {
            DPRINTK("Current Status is: CHECKING_FOR_MSR_SS_06_ACK\n");
            DPRINTK("    Command 0xE5 returns: 0x%02x\n", scancode);

            switch (scancode) {
            /* A missing Ack is the `broken` behaviour the SureOne keyboard
             * controller.
             *
             * If MSR is active it sends a Nack 0xFE plus a bad status byte
             * 0x55.
             *
             * Linux 2.4:
             *     The bad status means that the scancode is ignored by linux in pc_keyb.c
             * (handle_kbd_event) and never reaches the driver.
             *
             * Linux 2.6:
             *     0xFE comes to this filter
             */
            
            case 0xFA:  /* Ack for Command E5, implies Source is keyboard */
                {
                    DPRINTK("    Switching Off CHECKING_FOR_MSR_SS_06_ACK\n");

                    pKbdData->fStatus &= ~CHECKING_FOR_MSR_SS_06_ACK;

                    DPRINTK("    Switching On CHECKING_SOURCE_KEYBOARD_06\n");

                    pKbdData->fStatus |= CHECKING_SOURCE_KEYBOARD_06;

                    DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                    *pScancode = scancode = 0;
                    return 1;
                }
            default:    /* No Ack for Command E5, implies Source is MSR */
                {
                    /* Set SureOneMsrStatus collection stage */
                    if (pKbdData->SureOneMsrStatus.CollectionState == Waiting) {
                        DPRINTK("    Moving to Collection State: got_SS_1\n");

                        pKbdData->SureOneMsrStatus.CollectionState = got_SS_1;                  
                    } else if (pKbdData->SureOneMsrStatus.CollectionState ==  FirstTrackComplete) {
                        DPRINTK("    Moving to Collection State: got_SS_2\n");

                        pKbdData->SureOneMsrStatus.CollectionState = got_SS_2;

                        /* Switch Off the safety timer between
                         * FirstTrackComplete and SS 2
                         */
                        stop_inter_track_timer();
                    } else {
                        EPRINTK("    Unexpected Collection state 0x%02x\n",
                                pKbdData->SureOneMsrStatus.CollectionState);

                        /* Discard the buffer and Deactivate
                         * This also switches Off all Flags and puts us back
                         * into Inactive state
                         */
                        EPRINTK("    Discarding MSR buffer and deactivating\n");

                        __discard_and_deactivate_active_data_buffer();

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }

                    DPRINTK("    Switching Off CHECKING_FOR_MSR_SS_06_ACK\n");

                    pKbdData->fStatus &= ~CHECKING_FOR_MSR_SS_06_ACK;

                    DPRINTK("    Switching On COLLECTING_MSR_DATA\n");              

                    pKbdData->fStatus |= COLLECTING_MSR_DATA;

                    /* Start safety timeout timer: SS->End of Track */
                    start_end_track_timer((unsigned long)pKbdData->SureOneMsrStatus.CollectionState);

                    /* Change scancode to start sentinel
                     *
                     * We want to save the sentinel as part of the MSR data
                     * for SureOne We will also interpret it as a shifted
                     * scancode 0x2a, 0x06 for '%' character ... So inject a
                     * leading 0x2a into the buffer first.  We expect break
                     * code 0xaa (for this 0x2a) on the next ISR cycle
                     */
                    scancode = collecting_msr_data(0x2a);
                    scancode = collecting_msr_data(0x06);

                    /* Also adjust externally visible shift state to reflect
                     * that a 0x2a arrived before the 0x06 start sentinel,
                     * which will never be matched by the arrival of a 0xaa
                     * shift break which is visible to the keyboard driver
                     * stack, because it will be collected, buffered and
                     * consumed on the *next* ISR cycle.
                     */

                    DPRINTK("<<<-  rc=0x00, scancode=0xaa\n");
                    *pScancode = scancode = 0xaa;
                    return 0;
                }
            }
        }

        else if (pKbdData->fStatus & CHECKING_SOURCE_KEYBOARD_06) {
            DPRINTK("Current Status is: CHECKING_SOURCE_KEYBOARD_06\n");
            DPRINTK("    Command 0xE5 Reply returns: 0x%02x\n", scancode);
            DPRINTK("    Switching Off CHECKING_SOURCE_KEYBOARD_06\n");

            pKbdData->fStatus &= ~CHECKING_SOURCE_KEYBOARD_06;

            /* Change scancode to '%' Keypress
             * We want to pass on the original key
             */
            scancode = 0x06;

            DPRINTK("<<<-  rc=0x00, scancode=0x06\n");
            *pScancode = scancode;
            return 0;
        }

        /* ///////////////
         * Track Status //
         * ///////////////
         */
        else if (pKbdData->fStatus & CHECKING_FOR_MSR_TRACK_STATUS_ACK) {
            DPRINTK("Current Status is Checking For MSR Track Status\n");
            DPRINTK("Current Status is: CHECKING_FOR_MSR_TRACK_STATUS_ACK\n");
            DPRINTK("    Command 0xE4 returns: 0x%02x\n", scancode);

            /* We got here because the Inter-Track timer timed out. This means
             * that there was probably only one track of data read by the MSR.
             *
             * As a result of entering the timeout function we sent an 0xE4
             * command to the Keyboard Controller: Read MSR data format and
             * error code. If there is no new MSR activity we should get a
             * sensible reply on the next ISR cycle telling us how many tracks
             * the MSR reader actually did read. If it really was just one
             * then we will have our data.
             *
             * A missing Ack is the `broken` behaviour the SureOne keyboard
             * controller.
             *
             * If MSR is active it sends a Nack 0xFE plus a bad status byte
             * 0x55.
             *
             * Linux 2.4:
             * The bad status means that the scancode is ignored by linux in
             * pc_keyb.c
             * (handle_kbd_event) and never reaches the driver.
             *
             * Linux 2.6:
             * 0xFE comes to this filter
             *
             * This implies MSR is active (again). In this case we cannot get
             * MSR data format information. We must treat this as an error and
             * make sure the driver is returned to known state:
             * Not COLLECTING_MSR_DATA and 'Inactive'. Client must poll again
             * to put the driver back into the 'Waiting for start of MSR
             * Data' mode.
             */

            switch (scancode) {
            case 0xFA: /* Ack for Command E4, implies MSR is quiescent and the reply is good */
                {
                    DPRINTK("    Switching Off CHECKING_FOR_MSR_TRACK_STATUS_ACK\n");       

                    pKbdData->fStatus &= ~CHECKING_FOR_MSR_TRACK_STATUS_ACK;

                    DPRINTK("    Switching On CHECKING_MSR_TRACK_STATUS\n");

                    pKbdData->fStatus |= CHECKING_MSR_TRACK_STATUS;

                    DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                    *pScancode = scancode = 0;
                    return 1;
                }
            default: /* No Ack for Command E4, implies MSR is Active (Again?) */
                {
                    /* Discard the buffer and Deactivate
                     * This also switches Off all Flags and puts us back into
                     * Inactive state
                     */
                    EPRINTK("    Expecting Single track - but MSR is still active\n");
                    EPRINTK("    Discarding MSR buffer and deactivating\n");

                    __discard_and_deactivate_active_data_buffer();

                    DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                    *pScancode = scancode = 0;
                    return 1;
                }
            }
        }

        else if (pKbdData->fStatus & CHECKING_MSR_TRACK_STATUS) {
            DPRINTK("Current Status is: CHECKING_FOR_MSR_TRACK_STATUS\n");

            if (is_valid_single_track_data(scancode)) {
                DPRINTK("    is valid single track data: TRUE\n");

                DPRINTK("    Switching Off CHECKING_MSR_TRACK_STATUS\n");

                pKbdData->fStatus &= ~CHECKING_MSR_TRACK_STATUS;

                /* /////////////////////////////////
                 * Got a full buffer of MSR data //
                 * /////////////////////////////////
                 */

                /* Switch SureOne MSR Status to 'Holding MSR Data'
                 * This locks the buffer
                 */
                DPRINTK("    Moving to Collection State: Holding\n");

                pKbdData->SureOneMsrStatus.CollectionState = Holding;                  
            } else {
                DPRINTK("    is valid single trackdata: FALSE\n");

                /* Discard the buffer
                 * This also switches Off all Flags and puts us back into
                 * Waiting state
                 */
                EPRINTK("    Expecting Single track - but MSR actually read two tracks\n");
                EPRINTK("    Discarding MSR buffer and deactivating\n");

                __discard_and_deactivate_active_data_buffer();
            }

            DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
            *pScancode = scancode = 0;
            return 1;
        }

        /* /////////////////////////
         * Normal MSR Collecting //
         * /////////////////////////
         */
        else if (pKbdData->fStatus & COLLECTING_MSR_DATA) {
            DPRINTK("Current Status is: COLLECTING_MSR_DATA\n");
            DPRINTK("    Collection State: 0x%02x\n",
                    pKbdData->SureOneMsrStatus.CollectionState);

            /* ////////////
             * Got SS 1 //
             * ////////////
             */
            if (pKbdData->SureOneMsrStatus.CollectionState == got_SS_1) {
                if (SUREONE_SCANCODE_END_MSR_DATA == scancode) {    /* 0x35 */
                    DPRINTK("    GOT Potential ES, 0x35\n");

                    /* Only Shifted 0x35 is End Sentinel */

                    if ((pKbdData->StreamBuffer[1] == 0x2a) &&
                        (pKbdData->StreamBuffer[0] == 0x35)) {
                        DPRINTK("    Moving to Collection State: got_ES_1\n");

                        pKbdData->SureOneMsrStatus.CollectionState = got_ES_1;
                    } else {
                        /* else it is '/' as part of the msr data */
                        DPRINTK("    Ignoring\n");
                    }
                }

                scancode = collecting_msr_data(scancode);

                DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                *pScancode = scancode = 0;
                return 1;
            }

            /* ////////////
             * Got ES 1 //
             * ////////////
             */
            else if (pKbdData->SureOneMsrStatus.CollectionState == got_ES_1) {
                /* We are waiting for a sequence of 0xaa, 0x1c between the
                 * end of first track and the start of the next. Ignore
                 * scancodes that do not fit the pattern.
                 */

                DPRINTK("Collection State = got_ES_1, scancode= 0x%02x\n",scancode);
                switch (scancode) {
                case 0xaa:
                    {
                        scancode = collecting_msr_data(scancode);

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }
                case 0x1c:
                    {
                        scancode = collecting_msr_data(scancode);

                        /* Switch SureOne MSR Status to 'Got 0x1C between
                         * first and second tracks'  This will allow us to
                         * look for start of the next track.
                         */
                        DPRINTK("    Moving to Collection State: FirstTrackComplete\n");

                        pKbdData->SureOneMsrStatus.CollectionState = FirstTrackComplete;

                        /* Switch Off first track collect phase */
                        DPRINTK("    Switching Off COLLECTING_MSR_DATA\n");

                        pKbdData->fStatus &= ~COLLECTING_MSR_DATA;

                        /* Switch Off the safety timer between SS 1 and
                         * FirstTrackComplete
                         */
                        stop_end_track_timer();

                        /* Switch On the safety timer between
                         * FirstTrackComplete and SS 2
                         */
                        start_inter_track_timer();

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }
                default: /* Report, consume but do not collect */
                    {
                        EPRINTK("    Unexpected scancode 0x%02x while waiting for 0x1C between tracks - consumed\n", scancode);

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }
                }
            }

            /* ////////////
             * Got SS 2 //
             * ////////////
             */
            else if (pKbdData->SureOneMsrStatus.CollectionState == got_SS_2) {
                if (SUREONE_SCANCODE_END_MSR_DATA == scancode) {    /* 0x35 */
                    DPRINTK("    GOT Potential ES, 0x35\n");

                    /* Only Shifted 0x35 is End Sentinel */

                    if ((pKbdData->StreamBuffer[1] == 0x2a) &&
                        (pKbdData->StreamBuffer[0] == 0x35)) {
                        /* Move to next colection state - got ES 2 */
                        DPRINTK("    Moving to Collection State: got_ES_2\n");

                        pKbdData->SureOneMsrStatus.CollectionState = got_ES_2;
                    } else {
                        /* else it is '/' as part of the msr data */
                        DPRINTK("    Ignoring\n");
                    }
                }

                scancode = collecting_msr_data(scancode);

                DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                *pScancode = scancode = 0;
                return 1;
            }

            /* ///////////
             * Got ES 2 //
             * ///////////
             */
            else if (pKbdData->SureOneMsrStatus.CollectionState == got_ES_2) {
                /* We are waiting for a sequence of 0xaa, 0x1c at the end of
                 * second track Ignore scancodes that do not fit the pattern.
                 */

                switch (scancode) {
                case 0xaa:
                    {
                        scancode = collecting_msr_data(scancode);

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }
                case 0x1c:
                    {
                        scancode = collecting_msr_data(scancode);

                        /* ////////////////////////////////
                         * Got a full buffer of MSR data //
                         * ////////////////////////////////
                         */

                        /* Switch SureOne MSR Status to 'Holding MSR Data'
                         * This locks the buffer
                         */
                        DPRINTK("    Moving to Collection State: Holding\n");

                        pKbdData->SureOneMsrStatus.CollectionState = Holding;

                        /* Switch Off second track collect phase */
                        DPRINTK("    Switching Off COLLECTING_MSR_DATA\n");

                        pKbdData->fStatus &= ~COLLECTING_MSR_DATA;

                        /* Switch off the safety timer between SS 2 and
                         * Holding
                         */
                        stop_end_track_timer();

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }
                default: /* Report, consume but do not collect */
                    {
                        EPRINTK("    Unexpected scancode 0x%02x while waiting for 0x1C between tracks - consumed\n", scancode);

                        DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                        *pScancode = scancode = 0;
                        return 1;
                    }
                }
            } else {     /* Unexpected state */
                EPRINTK("    Unexpected Collection State: 0x%02x\n",
                        pKbdData->SureOneMsrStatus.CollectionState);

                /* Discard the buffer and Deactivate
                 * This also switches Off all Flags and puts us back into
                 * Inactive state
                 */
                EPRINTK("    Discarding MSR buffer and deactivating\n");

                __discard_and_deactivate_active_data_buffer();

                /* Early exit - Handled */
                DPRINTK("<<<-  rc=0x01, scancode=0x00\n");
                *pScancode = scancode = 0;
                return 1;
            }
        }

        /* /////////////////////////////////////////////////////////////////
         * At this point we have a scancode that we haven't stuffed into
         * a buffer 
         * /////////////////////////////////////////////////////////////////
         */

        switch (scancode) {
        case SUREONE_SCANCODE_BEG_MSR_DATA_27:    /* 0x27 */
            {
                SUREONE_MSR_COLLECTION_STATE msrState =
                pKbdData->SureOneMsrStatus.CollectionState;

                DPRINTK("GOT Possible SS 0x27: Collection state=0x%02x\n",
                        msrState);

                int ValidStartSequence = FALSE;

                if (msrState == Waiting) {
                    /* Look for First Track scancode sequence 0x2a 0xaa 0x27 */
                    DPRINTK("msrState waiting...FIRST PASS Streambuffer 0x%02x 0x%02x 0x%02x\n",
                            pKbdData->StreamBuffer[0],
                            pKbdData->StreamBuffer[1],
                            pKbdData->StreamBuffer[2]);

                    if ((pKbdData->StreamBuffer[2] == 0x2a) &&
                        (pKbdData->StreamBuffer[1] == 0xaa) &&
                        (pKbdData->StreamBuffer[0] == 0x27)) {
                        ValidStartSequence = TRUE;
                        DPRINTK("    Valid First Track Start Sequence - Will check if it is from MSR\n");
                    }
                } else if (msrState == FirstTrackComplete) {
                    /* Look for Second Track start sequence 0xaa 0x1c 0x27 */
                    DPRINTK("MSR SECOND TRACK SEQUENCE\n");
                    DPRINTK("pKbdData Stream buffer 0x%02x  0x%02x 0x%02x",
                            pKbdData->StreamBuffer[0],
                            pKbdData->StreamBuffer[1],
                            pKbdData->StreamBuffer[2]);

                    if ((pKbdData->StreamBuffer[2] == 0xaa) &&
                        (pKbdData->StreamBuffer[1] == 0x1c) &&
                        (pKbdData->StreamBuffer[0] == 0x27)) {
                        ValidStartSequence = TRUE;
                        DPRINTK("    Valid Second Track Start Sequence - Will check if it is from MSR\n");
                    }
                }

                if (ValidStartSequence == TRUE) {
                    /* Maybe Start of MSR Track
                     * Ask keyboard controller if data source is MSR
                     */

                    unsigned char status = inb(0x64);

                    while (status & 0x03) {
                        status = inb(0x64);
                    }

                    outb(0xE5, 0x60);

                    DPRINTK("    SENT 0xE5, status=0x%02x\n", status);
                    DPRINTK("    Switching On CHECKING_FOR_MSR_SS_27_ACK\n");

                    pKbdData->fStatus |= CHECKING_FOR_MSR_SS_27_ACK;
                    scancode = 0;
                }
#ifdef DEBUG
                else {
                    DPRINTK("    Ignoring\n");
                }
#endif

                break;
            }

        case SUREONE_SCANCODE_BEG_MSR_DATA_06:    /* 0x06 */
            {
                SUREONE_MSR_COLLECTION_STATE msrState =
                pKbdData->SureOneMsrStatus.CollectionState;

                DPRINTK("GOT Possible SS 0x06: Collection state=0x%02x\n", msrState);

                int ValidStartSequence = FALSE;

                if (msrState == Waiting) {
                    /* 0x06 is only a valid start sentinel on track #1 of 3
                     * Look for First Track scancode sequence 0x2a 0xaa 0x2a
                     * 0x06
                     */

                    DPRINTK("msrState waiting... ACTUAL STREAM BUFFER CONTENT: 0x%02x 0x%02x 0x%02x 0x%02x\n",pKbdData->StreamBuffer[0], pKbdData->StreamBuffer[1], pKbdData->StreamBuffer[2], pKbdData->StreamBuffer[3]);

                    if ((pKbdData->StreamBuffer[3] == 0x2a) &&
                        (pKbdData->StreamBuffer[2] == 0xaa) &&
                        (pKbdData->StreamBuffer[1] == 0x2a) &&
                        (pKbdData->StreamBuffer[0] == 0x06)) {
                        ValidStartSequence = TRUE;
                        DPRINTK("    Valid First Track Start Sequence - Will check if it is from MSR\n");
                    }
                }

                if (ValidStartSequence == TRUE) {
                    /* Ask keyboard controller if data source is MSR */
                    unsigned char status = inb(0x64);

                    while (status & 0x03) {
                        status = inb(0x64);
                    }

                    outb(0xE5, 0x60);

                    DPRINTK("    SENT 0xE5, status=0x%02x\n", status);
                    DPRINTK("    Switching On CHECKING_FOR_MSR_SS_06_ACK\n");

                    pKbdData->fStatus |= CHECKING_FOR_MSR_SS_06_ACK;
                    scancode = 0;
                }
#ifdef DEBUG
                else {
                    DPRINTK("    Ignoring\n");
                }
#endif

                break;
            }

        default:
            break;
        }
    }

    /* /////////////////////////////////////
     * Setup rc and scancode to be passed //
     * /////////////////////////////////////
     */

    if (scancode == 0) {
        /* Scancode is consumed by us */
        *pScancode = 0;
        rc = 1;
    } else {
        /* Return scancode for further processing */
        *pScancode = scancode;
        rc = 0;
    }

    DPRINTK("<<<-  rc=0x%02x, scancode=0x%02x\n", rc, *pScancode);

    return rc;
}


/************************************************************************
 *
 * Function Name:  ioctl
 *
 * Purpose:        Receive client requests
 *
 * Description:    Called when client requests service
 *
 *
 * Output:        Specific to the request and the struct passed down
 *                from the client.
 *
 *                Return 
 *                    0         - SUCCESS if service is supported
 *                    EINVAL    - Not supported 
 *
 * ***********************************************************************
 */

static int ioctl(struct inode *inode, struct file *fp, unsigned int cmd,
                  unsigned long userBuffer)
{
    unsigned long         CompletionCode;
    const unsigned long   ZeroLength = 0;

    GET_SLOT_INFO_PARMS   SlotInfo;
    QUERY_KEYBOARD_PARMS  KeyboardInfo;
    COMPLETION_PARMS_6    CompletionParms6;
    COMPLETION_PARMS_8    CompletionParms8;
    char fakeData[] = { ';','3','8','0','8','0','1','3','9','5','1','?', 13};

    int rc = SUCCESS;

    switch (cmd) {
#if defined(DEVELOPMENT)
    case FN_DEVELOPMENT_READ_RAW_MSR_DATA:
        {
            DPRINTK(">> ioctl(): FN_DEVELOPMENT_READ_RAW_MSR_DATA\n");

            if (pKbdData->MSRData.MSRDataLength &&
                (pKbdData->SureOneMsrStatus.CollectionState == Holding)) {
                DPRINTK("ioctl: reading MSR data %ld raw bytes\n", pKbdData->MSRData.MSRDataLength);

                copy_to_user (((PREAD_MSR_DATA_PARMS) userBuffer)->buffer,
                              &pKbdData->MSRData.MSRDataBuffer,
                              pKbdData->MSRData.MSRDataLength);

                copy_to_user(&(((PREAD_MSR_DATA_PARMS) userBuffer)->length),
                             &pKbdData->MSRData.MSRDataLength,
                             sizeof(pKbdData->MSRData.MSRDataLength));

                DPRINTK("ioctl(): MSR Data Buffer 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, ...\n",
                        pKbdData->MSRData.MSRDataBuffer[0],
                        pKbdData->MSRData.MSRDataBuffer[1],
                        pKbdData->MSRData.MSRDataBuffer[2],
                        pKbdData->MSRData.MSRDataBuffer[3],
                        pKbdData->MSRData.MSRDataBuffer[4],
                        pKbdData->MSRData.MSRDataBuffer[5],
                        pKbdData->MSRData.MSRDataBuffer[6]);

                /* Reset MSR Data:  Flag, Index, MSRDataLength &&
                 * MSRDataBuffer[]
                 */
                memset( &pKbdData->MSRData, 0, sizeof(pKbdData->MSRData));

                /* Set SureOneMsrStatus collection stage to Inactive */
                pKbdData->SureOneMsrStatus.CollectionState = Inactive;

            } else {

                DPRINTK("------  No Data, Yet  -----\n");

                copy_to_user(&((PREAD_MSR_DATA_PARMS) userBuffer)->length,
                             &ZeroLength,
                             sizeof(ZeroLength));

                /* Set SureOneMsrStatus collection stage to Waiting for new
                 * MSR event, if necessary
                 */
                if (pKbdData->SureOneMsrStatus.CollectionState < Waiting) {
                    pKbdData->SureOneMsrStatus.CollectionState = Waiting;
                }
            }

            rc = SUCCESS;

            CompletionCode = RPDONE;

            copy_to_user(&((PREAD_MSR_DATA_PARMS) userBuffer)->CompletionCode,
                         &CompletionCode,
                         sizeof(CompletionCode));

            DPRINTK("<< ioctl(): FN_DEVELOPMENT_READ_RAW_MSR_DATA\n");

            break;
        }
#endif

    case FN_READ_MSR_DATA:
        {
            DPRINTK(">> ioctl(): FN_READ_MSR_DATA\n");

            if ( pKbdData->MSRData.MSRDataLength &&
                ( (machine_type == IBM_4614 && 
                   pKbdData->SureOneMsrStatus.CollectionState == Holding) || 
                 (machine_type == IBM_4613 && 
                  pKbdData->SureOneMsrStatus.CollectionState == Inactive))) {
                DPRINTK("ioctl: reading MSR data %ld raw bytes\n",
                        pKbdData->MSRData.MSRDataLength);
                convert_raw_msr_data();

                copy_to_user (((PREAD_MSR_DATA_PARMS) userBuffer)->buffer,
                              &AsciiBuffer.DataBuffer,
                              AsciiBuffer.DataLength);

                copy_to_user(&(((PREAD_MSR_DATA_PARMS) userBuffer)->length),
                             &AsciiBuffer.DataLength,
                             sizeof(AsciiBuffer.DataLength));

                DPRINTK("ioctl(): MSR Ascii Buffer 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, ...\n",
                        AsciiBuffer.DataBuffer[0],
                        AsciiBuffer.DataBuffer[1],
                        AsciiBuffer.DataBuffer[2],
                        AsciiBuffer.DataBuffer[3],
                        AsciiBuffer.DataBuffer[4],
                        AsciiBuffer.DataBuffer[5],
                        AsciiBuffer.DataBuffer[6]);

                /* Reset MSR Data:  Flag, Index, MSRDataLength &&
                 * MSRDataBuffer[]
                 */
                memset( &pKbdData->MSRData, 0, sizeof(pKbdData->MSRData));

                /* Set SureOneMsrStatus collection stage to Inactive */
                pKbdData->SureOneMsrStatus.CollectionState = Inactive;

            } else {
                DPRINTK("    No Data, Yet!\n");

                copy_to_user(&((PREAD_MSR_DATA_PARMS) userBuffer)->length,
                             &ZeroLength,
                             sizeof(ZeroLength));

                /* Set SureOneMsrStatus collection stage to Waiting for new
                 * MSR event, if necessary
                 */
                if ( machine_type == IBM_4614 && pKbdData->SureOneMsrStatus.CollectionState < Waiting) {
                    pKbdData->SureOneMsrStatus.CollectionState = Waiting;
                }
            }
            rc = SUCCESS;
#if 0
	    memcpy (&AsciiBuffer.DataBuffer, fakeData, sizeof (fakeData));
	    AsciiBuffer.DataLength = sizeof (fakeData);

                copy_to_user (((PREAD_MSR_DATA_PARMS) userBuffer)->buffer,
                              &AsciiBuffer.DataBuffer,
                              AsciiBuffer.DataLength);

                copy_to_user(&(((PREAD_MSR_DATA_PARMS) userBuffer)->length),
                             &AsciiBuffer.DataLength,
                             sizeof(AsciiBuffer.DataLength));
#endif

            CompletionCode = RPDONE;

	    copy_to_user(&((PREAD_MSR_DATA_PARMS) userBuffer)->CompletionCode,
                         &CompletionCode,
                         sizeof(CompletionCode));

            DPRINTK("<< ioctl(): FN_READ_MSR_DATA\n");
            break;
        }

    case FN_CANCEL_MSR_READ:
        {
            DPRINTK(">> ioctl(): FN_CANCEL_MSR_READ\n");

            if (pKbdData->SureOneMsrStatus.CollectionState > Inactive) {
                discard_and_deactivate_active_data_buffer();
            }

            rc = SUCCESS;

            CompletionCode = RPDONE;

            copy_to_user(&((PREAD_MSR_DATA_PARMS) userBuffer)->CompletionCode,
                         &CompletionCode,
                         sizeof(CompletionCode));

            copy_to_user(&(((PREAD_MSR_DATA_PARMS) userBuffer)->length),
                         &ZeroLength,
                         sizeof(ZeroLength));

            DPRINTK("<< ioctl(): FN_CANCEL_MSR_READ\n");

            break;
        }

    case FN_GET_SLOT_INFO:
        {
            DPRINTK(">> ioctl(): FN_GET_SLOT_INFO\n");

            PGET_SLOT_INFO_PARMS pSlotInfo = &SlotInfo;

            memset(pSlotInfo, 0, sizeof(GET_SLOT_INFO_PARMS));

            pSlotInfo->CompletionCode   = RPDONE;
            pSlotInfo->AdapterID[1]     = 4614;
            if (machine_type == IBM_4613) {
                pSlotInfo->AdapterID[1] = 4613;
            }
            pSlotInfo->SlotDrawers[1]   = 0;
            pSlotInfo->SlotNVRAM[1]     = 0;

            copy_to_user((PGET_SLOT_INFO_PARMS) userBuffer,
                         pSlotInfo,
                         sizeof(GET_SLOT_INFO_PARMS));

            rc = SUCCESS;

            DPRINTK("<< ioctl(): FN_GET_SLOT_INFO\n");

            break;
        }

    case FN_QUERY_KEYBOARD:
        {
            DPRINTK(">> ioctl(): FN_QUERY_KEYBOARD\n");

            PQUERY_KEYBOARD_PARMS pKeyboardInfo = &KeyboardInfo;

            memset(pKeyboardInfo, 0, sizeof(QUERY_KEYBOARD_PARMS));

            pKeyboardInfo->CompletionCode   = RPDONE;
            pKeyboardInfo->msrType          = 1;
            pKeyboardInfo->keyboardEC       = pKbdData->EC_Level;  /* Kbd Controller EC */
            pKeyboardInfo->msrEC            = pKbdData->EC_Level;  /* Kbd Controller EC */
            pKeyboardInfo->keyboardID       = 4614;
            pKeyboardInfo->keyboardSubtype  = 5;                   /* SureOne Keyboard */
            if (machine_type == IBM_4613) {
                pKeyboardInfo->keyboardID       = 4613;
                pKeyboardInfo->keyboardSubtype  = 6;               /* IBM4613 Keyboard */
            }

            copy_to_user((PQUERY_KEYBOARD_PARMS) userBuffer,
                         pKeyboardInfo,
                         sizeof(QUERY_KEYBOARD_PARMS));

            rc = SUCCESS;

            DPRINTK("<< ioctl(): FN_QUERY_KEYBOARD\n");

            break;
        }

    case FN_CLICK:
    case FN_TONE:
    case FN_TRAP_KEYS:
    case FN_TRACKS:
    case FN_TYPEMATIC:
    case FN_SET_DOUBLEKEYS:
    case FN_POS_LEDS:
    case FN_ENABLE_DISABLE:
    case FN_GET_SHIFT_STATE:
    case FN_SET_SHIFT_STATE:
    case FN_QUERY_CODE_UPDATE:
        {
            /*DPRINTK(">> ioctl(): SUPPORTED BUT UNIMPLEMENTED FN: 0x%02x\n",cmd);*/

            PCOMPLETION_PARMS_6 pCompletionParms6 = &CompletionParms6;

            pCompletionParms6->CompletionCode = RPDONE;
            pCompletionParms6->length = 0;

            copy_to_user((PCOMPLETION_PARMS_6) userBuffer,
                         pCompletionParms6,
                         sizeof(COMPLETION_PARMS_6));

            rc = SUCCESS;

            /*DPRINTK("<< ioctl(): SUPPORTED BUT UNIMPLEMENTED FN: 0x%02x\n",cmd);*/

            break;
        }
    case FN_READ_KBD_STATUS:
    case FN_CODE_UPDATE:
        {
            /*DPRINTK(">> ioctl(): SUPPORTED BUT UNIMPLEMENTED FN: 0x%02x\n",cmd);*/

            PCOMPLETION_PARMS_8 pCompletionParms8 = &CompletionParms8;

            pCompletionParms8->CompletionCode = RPDONE;
            pCompletionParms8->length = 0;

            copy_to_user((PCOMPLETION_PARMS_8) userBuffer,
                         pCompletionParms8,
                         sizeof(COMPLETION_PARMS_8));

            rc = SUCCESS;

            /*DPRINTK("<< ioctl(): SUPPORTED BUT UNIMPLEMENTED FN: 0x%02x\n",cmd);*/

            break;
        }

    default:
        {
            EPRINTK("<< ioctl(): Invalid IOCTL, CMD=%d\n", cmd);
            rc = -EINVAL;
        }
    }

    return rc;
}


/************************************************************************
 *
 * Function Name:  check_model
 *
 * Purpose:        Read model number from bios.
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:          This not the SMBios value but the IBM model number at a
 *                 known location in the SureOne Bios
 *
 * ***********************************************************************
 */

static int check_model(void)
{
    void* page = (unsigned char*) ioremap(IOMAP_BASE, IOMAP_SIZE);

    if (!page)
        return FALSE;

    model = readl((unsigned char*)page + MODEL_OFFSET);
    iounmap((unsigned char*)page);

    DPRINTK("\nmodel -- 0x%04x\n",model);
    return TRUE;
}


/************************************************************************
 *
 * Function Name:  msr_init
 *
 * Purpose:        Initialise MSR state variables and objects
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called once on driver init
 *
 * ***********************************************************************
 */

static void msr_init(void)
{
    PSUREONE_MSR_STATUS pSureOneMsrStatus;

    pSureOneMsrStatus = &(pKbdData->SureOneMsrStatus);

    pSureOneMsrStatus->CollectionState = Inactive;

    init_timer(&(pSureOneMsrStatus->EndTrackTimer));
    init_timer(&(pSureOneMsrStatus->InterTrackTimer));
}


/************************************************************************
 *
 * Function Name:  reset_msr_status
 *
 * Purpose:        Reset MSR state variables and objects to inactive.
 *
 *
 * Output:
 *                 none
 *
 * ***********************************************************************
 */

static void reset_msr_status(void)
{
    PSUREONE_MSR_STATUS         pSureOneMsrStatus;

    pSureOneMsrStatus = &(pKbdData->SureOneMsrStatus);

    pSureOneMsrStatus->CollectionState = Inactive;

    del_timer_sync(&(pSureOneMsrStatus->EndTrackTimer));
    del_timer_sync(&(pSureOneMsrStatus->InterTrackTimer));
}


/************************************************************************
 *
 * Function Name:  update_stream_buffer
 *
 * Purpose:        Update the Stream Buffer
 *
 * Description:    Put the latest arrived scancode into the first buffer
 *                 position after shifting the current contents up one.
 *                 
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         The Stream Buffer is used to detect various states
 *                based on which scancode sequences have been received
 *
 * ***********************************************************************
 */

static inline void update_stream_buffer(unsigned char scancode)
{
    int i;

    for (i = STREAM_BUFFER_LAST_INDEX ; i > 0; i--) {
        pKbdData->StreamBuffer[i] = pKbdData->StreamBuffer[i-1];
    }

    pKbdData->StreamBuffer[0] = scancode;

    DPRINTK("StreamBuffer: 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, \n",
            pKbdData->StreamBuffer[0],
            pKbdData->StreamBuffer[1],
            pKbdData->StreamBuffer[2],
            pKbdData->StreamBuffer[3],
            pKbdData->StreamBuffer[4],
            pKbdData->StreamBuffer[5],
            pKbdData->StreamBuffer[6]);
}


/************************************************************************
 *
 * Function Name:  end_track_timed_out
 *
 * Purpose:        Notification that the end of track timer has timed out
 *
 * Description:    Notification that the end of track timer has timed out
 *                 called by the timer object
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Runaway MSR - a start of track was detected but no valid
 *                end was found within END_TRACK_TIMEOUT jiffies
 *
 * ***********************************************************************
 */

static void end_track_timed_out(unsigned long data) 
{
#ifdef DEBUG
    SUREONE_MSR_COLLECTION_STATE initialMsrState = (SUREONE_MSR_COLLECTION_STATE) data;

    DPRINTK("end-track timer timed out: jiffies=%lu, initial msrState=0x%02x\n",
            jiffies, initialMsrState);
#endif

    __discard_and_deactivate_active_data_buffer();
}


/************************************************************************
 *
 * Function Name:  inter_track_timed_out
 *
 * Purpose:        Notification that the inter-track timer has timed out
 *
 * Description:    Notification that the inter-track timer has timed out
 *                 called by the timer object
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:          Either only a single track was read by the MSR reader or
 *                 an error condition has been detected by the driver. To 
 *                 find out which we send command 0xE4 to the keyboard 
 *                 controller to determine how many tracks were read.
 *
 * ***********************************************************************
 */

static void inter_track_timed_out(unsigned long data)
{
    DPRINTK("inter-track timer timed out: jiffies=%lu\n", jiffies);

    unsigned char status = inb(0x64);

    while (status & 0x03) {
        status = inb(0x64);
    }

    outb(0xE4, 0x60);

    DPRINTK("    SENT 0xE4, status=0x%02x\n", status);

    /* Switch On Checking for MSR Track Status Ack */
    pKbdData->fStatus |= CHECKING_FOR_MSR_TRACK_STATUS_ACK;

    DPRINTK("    Switching On CHECKING_FOR_MSR_TRACK_STATUS_ACK\n");               
}


/************************************************************************
 *
 * Function Name:  start_end_track_timer
 *
 * Purpose:        Start the end of track timer
 *
 * Description:    Start the timer used to determine whether the MSR reader
 *                 has taken too long to read a single track.
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called when a valid end of track is detected. Used to 
 *                guard against runaway MSR - perhaps a damaged card.
 *
 * ***********************************************************************
 */

static void start_end_track_timer(unsigned long data)
{
    struct timer_list * pEndTrackTimer;

    pEndTrackTimer = &(pKbdData->SureOneMsrStatus.EndTrackTimer);

    if (machine_type == IBM_4613)
        pEndTrackTimer->expires = jiffies + END_TIMEOUT_BET_SCANCODES;
    else
        pEndTrackTimer->expires = jiffies + END_TRACK_TIMEOUT;

    pEndTrackTimer->function = end_track_timed_out;
    pEndTrackTimer->data = data;

    DPRINTK("start end-track timer: jiffies=%lu\n", jiffies);

    add_timer(pEndTrackTimer);
}


/************************************************************************
 *
 * Function Name:  stop_end_track_timer
 *
 * Purpose:        Stop the end of track timer
 *
 * Description:    Stop the timer that was started when an MSR Start 
 *                 Sentinel was detected
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called when a valid end of track is detected
 *
 * ***********************************************************************
 */

static void stop_end_track_timer(void)
{
    struct timer_list * pEndTrackTimer;

    pEndTrackTimer = &(pKbdData->SureOneMsrStatus.EndTrackTimer);

    DPRINTK("stop end-track timer: jiffies=%lu\n", jiffies);

    del_timer_sync(pEndTrackTimer);
}


/************************************************************************
 *
 * Function Name:  start_inter_track_timer
 *
 * Purpose:        Start the inter-track timer
 *
 * Description:    Start the timer used to determine whether the MSR reader
 *                 has read one or two tracks
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called when the end of the first track is detected
 *
 * ***********************************************************************
 */

static void start_inter_track_timer(void)
{
    struct timer_list * pInterTrackTimer;

    pInterTrackTimer = &(pKbdData->SureOneMsrStatus.InterTrackTimer);

    pInterTrackTimer->expires = jiffies + INTER_TRACK_TIMEOUT;
    pInterTrackTimer->function = inter_track_timed_out;
    pInterTrackTimer->data = (unsigned long) 0;

    DPRINTK("start inter-track timer: jiffies=%lu\n", jiffies);

    add_timer(pInterTrackTimer);
}


/************************************************************************
 *
 * Function Name:  stop_inter_track_timer
 *
 * Purpose:        Stop the inter-track timer
 *
 * Description:    Stop the timer that is started after a full first
 *                 track of MSR data was recieved.
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called when the start of a second track is detected
 *
 * ***********************************************************************
 */

static void stop_inter_track_timer(void)
{
    struct timer_list * pInterTrackTimer;

    pInterTrackTimer = &(pKbdData->SureOneMsrStatus.InterTrackTimer);

    DPRINTK("stop inter-track timer: jiffies=%lu\n", jiffies);

    del_timer_sync(pInterTrackTimer);
}


/************************************************************************
 *
 * Function Name:  is_valid_single_track_data
 *
 * Purpose:        Determine if a single MSR track was read by MSR reader.
 *
 * Description:    Sending command 0xE4 to the keyboard controller 
 *                 gets back a byte where bits 1, 2 and 3 show
 *                 whether tracks 1, 2 and/or 3 were read resp. This
 *                 function determines if a single track was read.
 *
 *
 * Output:
 *                 TRUE if only one track was read.
 *
 *
 * Notes:
 *
 * ***********************************************************************
 */

static int is_valid_single_track_data(unsigned char cmd_reply)
{
    int is_valid = FALSE;
    unsigned char reply         = (cmd_reply >> 1) & 0x07;
    unsigned char data_error    = (cmd_reply >> 4) & 0x01;

    if (data_error) {
        return FALSE;
    }

    switch (reply) {
    case 1: /* 001  Track 1 data received */
    case 2: /* 010  Track 2 data received */
    case 4: /* 100  Track 3 data received */
        {
            is_valid = TRUE;
            break;
        }
    }

    return is_valid;
}


/************************************************************************
 *
 * Function Name:  __discard_and_deactivate_active_data_buffer
 *
 * Purpose:        Discard current MSR buffer and return to a known state
 *
 * Description:    Clear MSR buffer, Resets all global state flags, moves
 *                 to MSR collection state: Inactive
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called from interrupt context when an unrecoverable 
 *                condition is detected during MSR collection
 *
 * ***********************************************************************
 */

static void __discard_and_deactivate_active_data_buffer(void)
{
    DPRINTK("discard and deactivate active data buffer:\n");

    /* Reset MSR Data:  Flag, Index, MSRDataLength && MSRDataBuffer[] */
    memset(&pKbdData->MSRData, 0, sizeof(pKbdData->MSRData));

    /* Reset SureOne MSR Status and delete any outstanding timers */
    reset_msr_status();

    /* Set SureOneMsrStatus collection stage to Inactive */
    pKbdData->SureOneMsrStatus.CollectionState = Inactive;

    DPRINTK("    Moving to collection state: Inactive\n");

    /* Switch Off All Flags */
    pKbdData->fStatus    = 0x00000000;

    DPRINTK("    Switching Off All State flags\n");               
}


/************************************************************************
 *
 * Function Name:  discard_and_deactivate_active_data_buffer
 *
 * Purpose:        Discard current MSR buffer and return to a known state
 *
 * Description:    Clear MSR buffer, Resets all global state flags, moves
 *                 to MSR collection state: Inactive
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called from Sys call context when the client wishes to
 *                cancel current MSR collection.
 *
 * ***********************************************************************
 */

static void discard_and_deactivate_active_data_buffer(void)
{
    unsigned long flags;

    local_irq_save(flags);
    __discard_and_deactivate_active_data_buffer();
    local_irq_restore(flags);
}


/************************************************************************
 *
 * Function Name:  convert_raw_msr_data
 *
 * Purpose:        Convert colected scancodes to Ascii characters
 *
 * Description:    Take the raw MSR scancodes collected in the MSR
 *                 data buffer and create Ascii characters for them
 *                 based on a mapping table in aipsokbps.h. The shift
 *                 state scancodes determine which mapping table is used.
 *
 *
 * Output:
 *                 none
 *
 *
 * Notes:         Called when a complete buffered swipe is about to be passed
 *                up to the client.
 *
 * ***********************************************************************
 */

static void convert_raw_msr_data()
{
    int i;
    int LShift = FALSE;
    PSUREONE_ASCII_BUFFER pAsciiBuffer = &AsciiBuffer;

    memset(pAsciiBuffer, 0, sizeof(SUREONE_ASCII_BUFFER));

    for (i = 0; i < pKbdData->MSRData.MSRDataLength; i++) {
        if (pAsciiBuffer->Flag == 1) {
            /* Buffer overrun, exit */
            break;
        }

        if (pKbdData->MSRData.MSRDataBuffer[i] == 0x2a) {
            LShift = TRUE;
        } else if (pKbdData->MSRData.MSRDataBuffer[i] == 0xaa) {
            LShift = FALSE;
        } else if (LShift) {
            pAsciiBuffer->DataBuffer[pAsciiBuffer->Index] =
            shift_kbd_map[pKbdData->MSRData.MSRDataBuffer[i]];

            pAsciiBuffer->Index++;
            pAsciiBuffer->DataLength++;
        } else if (!LShift) {
            pAsciiBuffer->DataBuffer[pAsciiBuffer->Index] =
            kbd_map[pKbdData->MSRData.MSRDataBuffer[i]];

            pAsciiBuffer->Index++;
            pAsciiBuffer->DataLength++;
        }

        /* Check Buffer */
        if (pAsciiBuffer->DataLength >= MAX_ASCII_BUFFER_SIZE) {
            EPRINTK("AsciiBuffer overrun at %lu\n",pAsciiBuffer->DataLength);
            pAsciiBuffer->Flag = 1;
        }
    }
}



/* /////////////////////
 * Load & Unload Fn`s //
 * /////////////////////
 */

module_init(sokb_init);
module_exit(sokb_exit);
